package universe
package animal

import scala.events._

import plant._

object Animal {
  
  def asAnimal(be:BoardElement): Option[Animal] = be match {
    case a: Animal => Some(a)
    case _ => None
  }

}

abstract class Animal(val world: World, var range:Int, val maxAge:Int) extends BoardElement {

  /** this event makes it possible to refine when an animal should move */
  evt movementTick[Unit] = world.clock

  evt born[Animal] = afterExec(appear) map ((_: Any) => this)
  
  evt disappeared[Animal] = afterExec(die) map ((_: Any) => this)

  /** Internal event to unregister everything that need to be unregistered before notifying the death to outer */
  protected evt shouldDie[Unit] = afterExec(get_older) && (_ => !canLive) || killed

  imperative evt killed[Unit]
  
  evt targetReached[BoardElement] = (afterExec(movement) && ((_: Any) => reached)) map ((_: Any) => target())

  protected evt newTargetNeeded[Unit] = targetReached || target.event(_.disappeared) || (movementTick && (() => target() == null))

  val isMale: Boolean

  var age = 0
  var energy = 15 
  val maxRange = 1 + range / 12

  // try to move to this position
  // we make use of the `Variable' class which facilitates the 
  // use of events of a mutable reference
  protected lazy val target: Variable[BoardElement] = Variable(null)

  observable def appear() {
    movementTick += movement _
    world.newDay += get_older _
    newTargetNeeded += nextTarget _
    shouldDie += die _
  }

  /** defines the next target and start moving to it */
  def nextTarget() {
    target := find_target(maxRange)
  }

  observable def die() {
    movementTick -= movement _
    world.newDay -= get_older _
    newTargetNeeded -= nextTarget _
    shouldDie -= die _
    board = null
  } 
   
  def isFertile: Boolean =
    age>10 && energy>40

  // check if the animal can still live
  def canLive = age < maxAge && energy > 0
  
  def get_older() {
    age+=1
    energy-=2 
  }

  // abstract method to find out where to go
  def find_target(range :Int):BoardElement 


  // try to move towards the target
  // for now allow jumping over
  def movement() {
    
    val targetpos = if (target()==null) { 
      pos.random_around(1+maxRange,1+maxRange*2.toInt) 
    } else { 
      target().pos
    }
    
    val delta=pos-targetpos
    var movement_x=0
    var movement_y=0
    var movement:Pos=null
    val steps=scala.math.sqrt(maxRange).toInt
    
    if (delta.x < 0) {
      movement_x=1 }
    if (delta.x > 0) {
      movement_x=0-1}
    if (delta.y < 0) {
      movement_y=1}
    if (delta.y > 0) {
      movement_y=0-1}
    if (movement_x*movement_x+movement_y*movement_y>1) {
      movement=new Pos(movement_x*steps,movement_y*steps)
    } else {
      movement=new Pos(movement_x*maxRange,movement_y*maxRange)
    }   
    
    val move_max=movement.distance
    val move_needed=delta.distance
    
    if (move_max>move_needed) {
      if (move_max>0) {
        movement=movement*(move_needed.toDouble/move_max)
      }
    }
    if (moveTo(pos+movement)) {
      energy -= energyLossFactor * movement.distance.toInt
    }
  }

  protected def energyLossFactor = 1/2

  def reached() =
    target() != null && distance(target()) < 2

  override def toString = "A(" + (energy.toFloat / 10) + ")#" + age + "@" + pos + "((" + super.toString + "))"

}
