package universe

import scala.collection.mutable.LinkedHashSet

import handler.AddedElementHandler

object Pos {
  lazy val rnd=new scala.util.Random
}

class Pos(val x :Int,val y:Int) {
  def +(p :Pos) = new Pos(p.x+x,p.y+y)

  def -(p :Pos) = new Pos(x-p.x,y-p.y)

  def *(f : Double) = new Pos((x*f).toInt,(y*f).toInt)

  override def equals(other: Any) : Boolean = other match {
    case that: Pos => that.x == this.x && that.y == this.y
    case _ => false
  }

  override def hashCode = (29 * x) + y

  def fast_distance(oth: Pos) = {
    val hor = (x-oth.x).abs
    val vert = (y-oth.y).abs
    if (hor > vert)
      hor
    else
      vert
  }

  def fast_distance = {
    val hor=x.abs
    val vert=y.abs
    if (hor>vert)
      hor
    else
      vert
  }
  
  def distance = scala.math.sqrt(x*x+y*y)
  
  def random_around(min:Int,max:Int) = {
    val d=min+Pos.rnd.nextInt(max-min)
    val xsign=if (Pos.rnd.nextBoolean) 1 else 0-1
    val ysign=if (Pos.rnd.nextBoolean) 1 else 0-1
    val xd=xsign*(min+Pos.rnd.nextInt(d))
    val yd=ysign*(0.5+min+scala.math.sqrt(d*d-xd*xd)).toInt
    new Pos(x+xd, y+yd)
  }

  override def toString = "["+x+","+y+"]"
}

trait BoardElement {  
  var pos = new Pos(0,0) 
  private var _board: Board = null
  private lazy val be_rnd=new scala.util.Random

  /** This event is triggered if this element disappears for any reason.
   *  Entities interested in this element as target must check that this
   *  element is still present when using it.
   */
  // begin disappeared @JVH
  val disappeared_subscribers=new scala.collection.mutable.ListBuffer[World] 

  def subscribe_disappeared(subscriber:World)={
    disappeared_subscribers+=subscriber
  }
  def unsubscribe_disappeared(subscriber:World)={
    disappeared_subscribers-=subscriber
  }
  def send_disappeared(be : BoardElement) ={
    disappeared_subscribers.foreach(_.handle_disappeared(be))
  }
  // end disappeared  @JVH
  // evt disappeared[BoardElement]   // @JVH

  def appear(): Unit

  def board_=(b: Board) {
    if (board == null || b == null) {
      if (b == null && board != null)
        board.remove(this)
      _board = b
    } 
  }

  def board = _board

  def x = pos.x
  
  def y = pos.y

  def move(r: Int): Boolean = {
    val delta=find_free_around(r)
    if (delta==null)
      false
    else if (!move(delta))
      move(r)
    else
      true
  }

  def find_free_around(r: Int, rmin:Int=1, maxDepth:Int=4) : Pos = {
    if (maxDepth<=0) {
      null
    } else if (board==null) {
      null
    } else { 
      val around=pos.random_around(rmin,r)
      if (board.free(around))
        around
      else
        find_free_around(r,rmin,maxDepth-1)
    }
  }
  
  
  def move(delta: Pos) = {
    if (board == null)
      false
    else
      board.moveTo(pos+delta,this)
  }

  def hasPos(oth : Pos) = {
    pos == oth
  }

  def moveTo(dst : Pos) : Boolean = {
    if (dst == null)
      false
    else if (hasPos(dst))
      true
    else if (board != null && !board.free(dst))
      false
    else {
      pos = dst
      true
    }
  }
  

  def fast_distance(b : BoardElement) = {
    if (board != b.board)
      Int.MaxValue
    else
      pos.fast_distance(b.pos)
  }

  def distance(b : BoardElement) = {
    if (board != b.board)
      Int.MaxValue
    else
      (pos-b.pos).distance.toInt
  }

  def symbol: String

  override def toString = this.getClass + "@" + posString

  def posString = pos.toString
}

class Board(val width:Int,val height:Int)  extends AddedElementHandler {
  val surface = new LinkedHashSet[BoardElement]
  // imperative evt addedElement[BoardElement]

  //  removedElement is not used it seems
  // evt removedElement[BoardElement] = afterExec(remove) map ((be: BoardElement, _: Unit) => be)

  // begin addedElement
  val addedElement_subscribers=new scala.collection.mutable.HashSet[AddedElementHandler] 
  def subscribe_addedElement(subscriber:AddedElementHandler)={
    addedElement_subscribers+=subscriber
  }
  def unsubscribe_addedElement(subscriber:AddedElementHandler)={
    addedElement_subscribers-=subscriber
  }
  def send_addedElement(be : BoardElement) ={
    addedElement_subscribers.foreach(_.handle_addedElement(be))
  }
  // end addedElement


  def onBoard(pos : Pos) : Boolean = {
    if (pos.x < 0 || pos.x >= width)
      false
    else if (pos.y<0 || pos.y>=height)
      false
    else
      true
  }

  def apply(x :Int, y:Int): BoardElement = this(new Pos(x,y))

  def apply(pos : Pos): BoardElement = { 
    if (!onBoard(pos)) {
      null
    } else {
      val fnd = surface.find(p => p.hasPos(pos))
      fnd match {
        case None => null
        case Some(x) => x
      }
    }
  }

  def free(pos : Pos) : Boolean = {
    if (!onBoard(pos))
      false
    else
      apply(pos) == null
  }

  def moveTo(pos: Pos, be:BoardElement) : Boolean = {
    if (!onBoard(pos))
      false
    else if (be.board == this && be.hasPos(pos))
      true
    else if (free(pos)) {
      if (be.board == null) {
        be.board = this
        surface += be
        send_addedElement(be)
      }
      be.moveTo(pos)
    } else
      false
  }

  // @JVH addedElement += (_.appear) 

  subscribe_addedElement(this) // @JVH

  // @JVH begin 
  def handle_addedElement(be : BoardElement) {
    be.appear    //   += (_.appear)
  }
  // @JVH end


    
  def neighbours(be:BoardElement,n:Int =0) : List[BoardElement] ={
    val full = surface.toList
    val selection = if (n==0) full else full.filter(p => be.fast_distance(p)<=n && be.distance(p)<=n) 

    val sorted = selection.sortWith((p1,p2) => be.distance(p1) < be.distance(p2))
    // consider only this board (so distance!=Int.MaxValue
    sorted.filter(p=> surface.contains(p))
  }
    
  def remove(be:BoardElement) {
    if (be.board == this || be.board == null)
      surface-=be
  }
  
}
